@@ -54,7 +54,7 @@ def test_fmt_string():
         pytest.param("a\n\n\xa0\n\nb"),  # lone NBSP between two paragraphs
         pytest.param("\xa0\n\n# heading"),  # lone NBSP followed by a heading
         pytest.param(
-            "```\na\n```\n\u2003\n# A\n", marks=pytest.mark.xfail()
+            "```\na\n```\n\u2003\n# A\n"
         ),  # em space surrounded by code and header
     ],
 )
@@ -141,3 +141,8 @@ def test_mdrenderer_no_finalize(tmp_path):
     unfinalized = MDRenderer().render(tokens, {}, env, finalize=False)
     finalized = MDRenderer().render(tokens, {}, env)
     assert finalized == unfinalized + "\n\n[gl ref]: https://gitlab.com\n"
+
+
+def test_ascii_whitespace_deprecation():
+    with pytest.warns(DeprecationWarning):
+        mdformat.codepoints.ASCII_WHITESPACE
