@@ -1,5 +1,4 @@
 import argparse
-import json
 from textwrap import dedent
 from unittest.mock import patch
 
@@ -8,16 +7,30 @@ import pytest
 
 import mdformat
 from mdformat._cli import run
-from mdformat.plugins import CODEFORMATTERS, PARSER_EXTENSIONS
-from mdformat.renderer import MDRenderer, RenderContext, RenderTreeNode
-
-
-def example_formatter(code, info):
-    return "dummy\n"
+from mdformat._compat import importlib_metadata
+from mdformat.plugins import (
+    _PARSER_EXTENSION_DISTS,
+    CODEFORMATTERS,
+    PARSER_EXTENSIONS,
+    _load_entrypoints,
+)
+from mdformat.renderer import MDRenderer
+from tests.utils import (
+    ASTChangingPlugin,
+    JSONFormatterPlugin,
+    PrefixPostprocessPlugin,
+    SuffixPostprocessPlugin,
+    TablePlugin,
+    TextEditorPlugin,
+    run_with_clear_cache,
+)
 
 
 def test_code_formatter(monkeypatch):
-    monkeypatch.setitem(CODEFORMATTERS, "lang", example_formatter)
+    def fmt_func(code, info):
+        return "dummy\n"
+
+    monkeypatch.setitem(CODEFORMATTERS, "lang", fmt_func)
     text = mdformat.text(
         dedent(
             """\
@@ -37,19 +50,80 @@ def test_code_formatter(monkeypatch):
     )
 
 
-class TextEditorPlugin:
-    """A plugin that makes all text the same."""
+def test_code_formatter__empty_str(monkeypatch):
+    def fmt_func(code, info):
+        return ""
 
-    @staticmethod
-    def update_mdit(mdit: MarkdownIt):
-        pass
+    monkeypatch.setitem(CODEFORMATTERS, "lang", fmt_func)
+    text = mdformat.text(
+        dedent(
+            """\
+    ~~~lang
+    aag
+    gw
+    ~~~
+    """
+        ),
+        codeformatters={"lang"},
+    )
+    assert text == dedent(
+        """\
+    ```lang
+    ```
+    """
+    )
 
-    def _text_renderer(  # type: ignore[misc]
-        tree: RenderTreeNode, context: RenderContext
-    ) -> str:
-        return "All text is like this now!"
 
-    RENDERERS = {"text": _text_renderer}
+def test_code_formatter__no_end_newline(monkeypatch):
+    def fmt_func(code, info):
+        return "dummy\ndum"
+
+    monkeypatch.setitem(CODEFORMATTERS, "lang", fmt_func)
+    text = mdformat.text(
+        dedent(
+            """\
+    ```lang
+    ```
+    """
+        ),
+        codeformatters={"lang"},
+    )
+    assert text == dedent(
+        """\
+    ```lang
+    dummy
+    dum
+    ```
+    """
+    )
+
+
+def test_code_formatter__interface(monkeypatch):
+    def fmt_func(code, info):
+        return info + code * 2
+
+    monkeypatch.setitem(CODEFORMATTERS, "lang", fmt_func)
+    text = mdformat.text(
+        dedent(
+            """\
+    ```    lang  long
+    multi
+    mul
+    ```
+    """
+        ),
+        codeformatters={"lang"},
+    )
+    assert text == dedent(
+        """\
+    ```lang  long
+    lang  longmulti
+    mul
+    multi
+    mul
+    ```
+    """
+    )
 
 
 def test_single_token_extension(monkeypatch):
@@ -75,24 +149,9 @@ def test_single_token_extension(monkeypatch):
     )
 
 
-class ExampleTablePlugin:
-    """A plugin that adds table extension to the parser."""
-
-    @staticmethod
-    def update_mdit(mdit: MarkdownIt):
-        mdit.enable("table")
-
-    def _table_renderer(  # type: ignore[misc]
-        tree: RenderTreeNode, context: RenderContext
-    ) -> str:
-        return "dummy 21"
-
-    RENDERERS = {"table": _table_renderer}
-
-
 def test_table(monkeypatch):
     """Test the table plugin, as a multi-token extension example."""
-    monkeypatch.setitem(PARSER_EXTENSIONS, "table", ExampleTablePlugin)
+    monkeypatch.setitem(PARSER_EXTENSIONS, "table", TablePlugin)
     text = mdformat.text(
         dedent(
             """\
@@ -147,27 +206,92 @@ def test_cli_options(monkeypatch, tmp_path):
     assert opts["mdformat"]["arg_name"] == 4
 
 
-class ExampleASTChangingPlugin:
-    """A plugin that makes AST breaking formatting changes."""
-
-    CHANGES_AST = True
-
-    TEXT_REPLACEMENT = "Content replaced completely. AST is now broken!"
+class ExamplePluginWithGroupedCli:
+    """A plugin that adds CLI options."""
 
     @staticmethod
     def update_mdit(mdit: MarkdownIt):
-        pass
+        mdit.enable("table")
 
-    def _text_renderer(  # type: ignore[misc]
-        tree: RenderTreeNode, context: RenderContext
-    ) -> str:
-        return ExampleASTChangingPlugin.TEXT_REPLACEMENT
+    @staticmethod
+    def add_cli_argument_group(group: argparse._ArgumentGroup) -> None:
+        group.add_argument("--o1", type=str)
+        group.add_argument("--o2", type=str, default="a")
+        group.add_argument("--o3", dest="arg_name", type=int)
+        group.add_argument("--override-toml")
+        group.add_argument("--dont-override-toml", action="store_const", const=True)
 
-    RENDERERS = {"text": _text_renderer}
+
+def test_cli_options_group(monkeypatch, tmp_path):
+    """Test that CLI arguments added by plugins are correctly added to the
+    options dict.
+
+    Use add_cli_argument_group plugin API.
+    """
+    monkeypatch.setitem(PARSER_EXTENSIONS, "table", ExamplePluginWithGroupedCli)
+    file_path = tmp_path / "test_markdown.md"
+    conf_path = tmp_path / ".mdformat.toml"
+    file_path.touch()
+    conf_path.write_text(
+        """\
+[plugin.table]
+override_toml = 'failed'
+toml_only = true
+dont_override_toml = 'dont override this with None if CLI opt not given'
+"""
+    )
+
+    with patch.object(MDRenderer, "render", return_value="") as mock_render:
+        assert (
+            run(
+                (
+                    str(file_path),
+                    "--o1",
+                    "other",
+                    "--o3",
+                    "4",
+                    "--override-toml",
+                    "success",
+                )
+            )
+            == 0
+        )
+
+    (call_,) = mock_render.call_args_list
+    posargs = call_[0]
+    # Options is the second positional arg of MDRender.render
+    opts = posargs[1]
+    table_opts = opts["mdformat"]["plugin"]["table"]
+    assert table_opts["o1"] == "other"
+    assert table_opts["o2"] == "a"
+    assert table_opts["arg_name"] == 4
+    assert table_opts["override_toml"] == "success"
+    assert table_opts["toml_only"] is True
+    assert (
+        table_opts["dont_override_toml"]
+        == "dont override this with None if CLI opt not given"
+    )
+
+
+def test_cli_options_group__no_toml(monkeypatch, tmp_path):
+    """Test add_cli_argument_group plugin API with configuration only from
+    CLI."""
+    monkeypatch.setitem(PARSER_EXTENSIONS, "table", ExamplePluginWithGroupedCli)
+    file_path = tmp_path / "test_markdown.md"
+    file_path.touch()
+
+    with patch.object(MDRenderer, "render", return_value="") as mock_render:
+        assert run((str(file_path), "--o1", "other")) == 0
+
+    (call_,) = mock_render.call_args_list
+    posargs = call_[0]
+    # Options is the second positional arg of MDRender.render
+    opts = posargs[1]
+    assert opts["mdformat"]["plugin"]["table"]["o1"] == "other"
 
 
 def test_ast_changing_plugin(monkeypatch, tmp_path):
-    plugin = ExampleASTChangingPlugin()
+    plugin = ASTChangingPlugin()
     monkeypatch.setitem(PARSER_EXTENSIONS, "ast_changer", plugin)
     file_path = tmp_path / "test_markdown.md"
 
@@ -186,15 +310,6 @@ def test_ast_changing_plugin(monkeypatch, tmp_path):
     assert file_path.read_text() == "Some markdown here\n"
 
 
-class JSONFormatterPlugin:
-    """A code formatter plugin that formats JSON."""
-
-    @staticmethod
-    def format_json(unformatted: str, _info_str: str) -> str:
-        parsed = json.loads(unformatted)
-        return json.dumps(parsed, indent=2) + "\n"
-
-
 def test_code_format_warnings__cli(monkeypatch, tmp_path, capsys):
     monkeypatch.setitem(CODEFORMATTERS, "json", JSONFormatterPlugin.format_json)
     file_path = tmp_path / "test_markdown.md"
@@ -224,7 +339,7 @@ def test_plugin_conflict(monkeypatch, tmp_path, capsys):
     plugin_name_1 = "plug1"
     plugin_name_2 = "plug2"
     monkeypatch.setitem(PARSER_EXTENSIONS, plugin_name_1, TextEditorPlugin)
-    monkeypatch.setitem(PARSER_EXTENSIONS, plugin_name_2, ExampleASTChangingPlugin)
+    monkeypatch.setitem(PARSER_EXTENSIONS, plugin_name_2, ASTChangingPlugin)
 
     file_path = tmp_path / "test_markdown.md"
     file_path.write_text("some markdown here")
@@ -237,49 +352,15 @@ def test_plugin_conflict(monkeypatch, tmp_path, capsys):
 
 
 def test_plugin_versions_in_cli_help(monkeypatch, capsys):
-    monkeypatch.setitem(PARSER_EXTENSIONS, "table", ExampleTablePlugin)
+    monkeypatch.setitem(
+        _PARSER_EXTENSION_DISTS, "table-dist", ("v3.2.1", ["table-ext"])
+    )
     with pytest.raises(SystemExit) as exc_info:
         run(["--help"])
     assert exc_info.value.code == 0
     captured = capsys.readouterr()
-    assert "Installed plugins:" in captured.out
-    assert "tests: unknown" in captured.out
-
-
-class PrefixPostprocessPlugin:
-    """A plugin that postprocesses text, adding a prefix."""
-
-    CHANGES_AST = True
-
-    @staticmethod
-    def update_mdit(mdit: MarkdownIt):
-        pass
-
-    def _text_postprocess(  # type: ignore[misc]
-        text: str, tree: RenderTreeNode, context: RenderContext
-    ) -> str:
-        return "Prefixed!" + text
-
-    RENDERERS: dict = {}
-    POSTPROCESSORS = {"text": _text_postprocess}
-
-
-class SuffixPostprocessPlugin:
-    """A plugin that postprocesses text, adding a suffix."""
-
-    CHANGES_AST = True
-
-    @staticmethod
-    def update_mdit(mdit: MarkdownIt):
-        pass
-
-    def _text_postprocess(  # type: ignore[misc]
-        text: str, tree: RenderTreeNode, context: RenderContext
-    ) -> str:
-        return text + "Suffixed!"
-
-    RENDERERS: dict = {}
-    POSTPROCESSORS = {"text": _text_postprocess}
+    assert "installed extensions:" in captured.out
+    assert "table-dist: table-ext" in captured.out
 
 
 def test_postprocess_plugins(monkeypatch):
@@ -305,3 +386,83 @@ def test_postprocess_plugins(monkeypatch):
         Prefixed!Example paragraph.Suffixed!
         """
     )
+
+
+def test_load_entrypoints(tmp_path, monkeypatch):
+    """Test the function that loads plugins to constants."""
+    # Create a minimal .dist-info to create EntryPoints out of
+    dist_info_path = tmp_path / "mdformat_gfm-0.3.6.dist-info"
+    dist_info_path.mkdir()
+    entry_points_path = dist_info_path / "entry_points.txt"
+    metadata_path = dist_info_path / "METADATA"
+    # The modules here will get loaded so use ones we know will always exist
+    # (even though they aren't actual extensions).
+    entry_points_path.write_text(
+        """\
+[mdformat.parser_extension]
+ext1=mdformat.plugins
+ext2=mdformat.plugins
+"""
+    )
+    metadata_path.write_text(
+        """\
+Metadata-Version: 2.1
+Name: mdformat-gfm
+Version: 0.3.6
+"""
+    )
+    distro = importlib_metadata.PathDistribution(dist_info_path)
+    entrypoints = distro.entry_points
+
+    loaded_eps, dist_infos = _load_entrypoints(entrypoints)
+    assert loaded_eps == {"ext1": mdformat.plugins, "ext2": mdformat.plugins}
+    assert dist_infos == {"mdformat-gfm": ("0.3.6", ["ext1", "ext2"])}
+
+
+def test_no_codeformatters__toml(tmp_path, monkeypatch):
+    monkeypatch.setitem(CODEFORMATTERS, "json", JSONFormatterPlugin.format_json)
+    unformatted = """\
+```json
+{"a": "b"}
+```
+"""
+    formatted = """\
+```json
+{
+  "a": "b"
+}
+```
+"""
+    file1_path = tmp_path / "file1.md"
+
+    # Without TOML
+    file1_path.write_text(unformatted)
+    assert run((str(tmp_path),)) == 0
+    assert file1_path.read_text() == formatted
+
+    # With TOML
+    file1_path.write_text(unformatted)
+    config_path = tmp_path / ".mdformat.toml"
+    config_path.write_text("codeformatters = []")
+    assert run_with_clear_cache((str(tmp_path),)) == 0
+    assert file1_path.read_text() == unformatted
+
+
+def test_no_extensions__toml(tmp_path, monkeypatch):
+    plugin = ASTChangingPlugin()
+    monkeypatch.setitem(PARSER_EXTENSIONS, "ast_changer", plugin)
+    unformatted = "text\n"
+    formatted = plugin.TEXT_REPLACEMENT + "\n"
+    file1_path = tmp_path / "file1.md"
+
+    # Without TOML
+    file1_path.write_text(unformatted)
+    assert run((str(tmp_path),)) == 0
+    assert file1_path.read_text() == formatted
+
+    # With TOML
+    file1_path.write_text(unformatted)
+    config_path = tmp_path / ".mdformat.toml"
+    config_path.write_text("extensions = []")
+    assert run_with_clear_cache((str(tmp_path),)) == 0
+    assert file1_path.read_text() == unformatted
